#
# Thatch - Tcl Attach
#
#
# Stuart Cassoff
#
# May 2009
# December 2009
# January 2010
# October 2010
# March 2012
# Summer 2012
# Spring 2018
# Version 0.1
#
# Spring 2018
# Version 0.2
#
# Summer 2018
# Version 0.3
#
# Summer 2022
# Version 0.4
#
# Summer 2022
# Version 0.5
#


[namespace eval ::thatch {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	thatch
		title	Thatch
		descr	"App Builder"
		author	{Stuart Cassoff}
		when	{Summer 2022}
		version	0.5
	}]
	dict set pkgconfig importfrom [namespace current]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2009,2018,2022 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}; # pkgconfig


#-------------------------------------------------------------------
# Thatch preamble and bootstrap code
#-------------------------------------------------------------------


proc preamble {{wish {0}}} {
	return "#! /bin/sh\n# \\\nexec [expr {$wish ? "wi" : "tcl"}]sh\${TCL} \"\$0\" \${1+\"\$@\"}"
}


variable bootstrap [string range {
#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------
namespace eval ::thatch {
	if {[package provide thatch] ne {}} { return }

	proc thatchBootstrap {{fn ""}} {
		if {$fn eq ""} { set fn [info script] }
		if {[catch {
			set f [open $fn r]
			fconfigure $f -eofchar \x1a
			read $f
			seek $f 1 current
			set d [read $f]
			close $f
		} err]} {
			return 0
		}
		if {[string length $d] == 0} { return 0 }
		uplevel #0 $d
		package require thatch
	}
	package ifneeded thatch @VER@ [namespace current]::thatchBootstrap
}

package require thatch
::thatch::thatchSetup [info script] @THATCHOPTS@

source @MAIN@
#-------------------------------------------------------------------
} 1 end-1]


#-------------------------------------------------------------------
# Other
#-------------------------------------------------------------------

proc linterleave {l1 l2} {
	set l {}
	foreach e1 $l1 e2 $l2 { lappend l $e1 $e2 }
	return $l
}

proc dgwd {def dict key args} {
	if {[dict exists $dict $key {*}$args]} {
		return [dict get $dict $key {*}$args]
	}
	return $def
}

# In My Namespace
proc imn {what} {
	return [string trimright [uplevel 1 {namespace current}] :]::$what
}

proc aliasconst {name val args} {
	if {[llength $args] % 2 != 0} { error "must be %2!" }
	foreach {n v} [linsert $args 0 $name $val] {
		interp alias {} [imn $n] {} return -level 0 $v
	}
}

proc unknownOption {opt opts} {
	set msg "unknown option \"$opt\""
	if {[llength $opts] == 0} { return $msg }
	append msg ", must be: "
	append msg [lindex $opts 0]
	if {[llength $opts] == 2} {
		append msg " or [lindex $opts 1]"
	} elseif {[llength $opts] > 2} {
		append msg ", " [join [lrange $opts 1 end-1] ", "] " or " [lindex $opts end]
	}
	return $msg
}

#-------------------------------------------------------------------
# File
#-------------------------------------------------------------------

proc openItIn {fn {mode r} args} {
	if {$fn in {- stdin}} {
		set f stdin
	} else {
		if {[string index $fn 0] eq "@"} {
			set f [string range $fn 1 end]
		} else {
			set f [open $fn $mode]
		}
	}
	if {[llength $args] > 0} {
		chan configure $f {*}$args
	}
	return $f
}

proc openItOut {fn {mode w} args} {
	if {$fn in {- stdout}} {
		set f stdout
	} else {
		set f [open $fn $mode]
	}
	if {[llength $args] > 0} {
		chan configure $f {*}$args
	}
	return $f
}

proc closeIt {f} {
	if {$f in {stdin stdout}} {
		return {}
	}
	return [chan close $f]
}

proc inhaleIt {fn args} {
	return [read [set f [openItIn $fn r {*}$args]]][closeIt $f]
}

interp alias {} [imn openIt] {} [imn openItIn]

#-------------------------------------------------------------------
# Transforms
#-------------------------------------------------------------------

proc thatchWhatTransforms {{pat *}} {
	return [lsort -dictionary [lsearch -inline -all -glob [list copy base64 zip] $pat]]
}

if {[namespace ensemble exists binary] && "encode" in [dict keys [namespace ensemble configure binary -map]] && "decode" in [dict keys [namespace ensemble configure binary -map]]} {
	namespace eval ::base64 {
		proc encode {args} { binary encode base64 {*}$args }
		proc decode {args} { binary decode base64 {*}$args }
		package provide base64 2000
	}
}; # if binary encode/decode

namespace eval ::zip {
	proc encode {args} { zlib compress   {*}$args }
	proc decode {args} { zlib decompress {*}$args }
	package provide zip 2000
}

proc thatchTransform {data process args} {
	foreach alg $args {
		if {$alg eq "copy"} { continue }
		package require $alg
		set data [::${alg}::${process} $data]
	}
	return $data
}

#-------------------------------------------------------------------
# Checksums
#-------------------------------------------------------------------

proc thatchWhatChecksums {{pat *}} {
	return [lsort -dictionary [lsearch -inline -all -glob [list md5 sha256] $pat]]
}

proc thatchChecksum {data alg} {
	package require $alg
	set ns [string map [list md5 md5 sha256 sha2] $alg]
	return [::${ns}::${alg} -hex $data]
}

proc thatchMakeChecksums {data args} {
	set cksums {}
	foreach alg [lsort -nocase -dictionary $args] {
		set alg [string tolower $alg]
		lappend cksums $alg:[thatchChecksum $data $alg]
	}
	return [join $cksums ,]
}

proc thatchCheckChecksums {data args} {
	set errs {}
	foreach cksum [lsort -nocase -dictionary [split $args ,]] {
		lassign [split $cksum :] alg sum
		set alg [string tolower $alg]
		set csum [thatchChecksum $data $alg]
		if {$csum ne $sum} {
			lappend errs [list $cksum $csum]
		}
	}
	return $errs
}

#-------------------------------------------------------------------
# Thatch
#-------------------------------------------------------------------

#-------------------------------------------------------------------
#
#	i == file info
#	N == \n
#	Z == ^Z
#
#	ZN# THATCH iiiiiiiiiiiiiiiN
#	||__sig___||__file info__||
#	|          |file info line|
#	|_____file info chunk_____|
#
#
#	file info       fis
#	file info line  fil
#	file info chunk fic
#	file info dict  fid
#	file spec       fs
#
#-------------------------------------------------------------------


#
# File Spec (fs) fields
#
proc fsFields {} {
	return [list name size mtime base64 zip md5 sha256 remove rehome modbase]
}

proc fid->fis {fid} {
	set name [dict get $fid name]
	if {[dict exists $fid rehome]} {
		if {[dict get $fid rehome] ne ""} {
			set name [dict get $fid rehome]
		}
		dict unset fid name
		dict unset fid rehome
	}
	if {[dict exists $fid modbase] && [dict get $fid modbase] eq {}} {
		dict unset fid modbase
	}
	return "$name $fid"
}

proc fis->fid {fis} {
	set q [dict create name [lindex $fis 0] {*}[lrange $fis 1 end]]
}

proc fidFields {args} {
	return "name transforms checksums size stored-size mtime modbase"
}

proc newFid {args} {
	return [dict create ${*}[fidFields] {*}$args]
}

#
# File Spec (fs) -> File Info Dict (fid)
#
proc fs->fid {fs {dv {}}} {
	if {$dv ne ""} { upvar $dv data }
	lassign $fs {*}[fsFields]
	set transforms [list]
	set chanconf [dict create]

	# If zip & base64 do zip first
	if {$zip == 1} {
		lappend transforms zip
		dict set chanconf -translation binary
	}
	if {$base64 == 1} {
		lappend transforms base64
		dict set chanconf -translation binary
	}

	set fid [dict create name $name]

	if {[llength $transforms] > 0} { dict set fid transforms $transforms }

	if {$dv eq ""} {
		return $fid
	}

	set filename $name
	file stat $filename s
	set data [inhaleIt $filename {*}$chanconf]

	set algs [list]
	if {$md5 == 1}  { lappend algs md5 }
	if {$sha256 == 1} { lappend algs sha256 }

	set checksums [thatchMakeChecksums $data {*}$algs]
	set data [thatchTransform $data encode {*}$transforms]
	set s(stored-size) [string bytelength $data]

	if {$checksums ne ""} {
		dict set fid checksums $checksums
	}

	if {$size == 1} {
		dict set fid size $s(size)
		if {$s(size) != $s(stored-size)} {
			dict set fid stored-size $s(stored-size)
		}
	}

	if {$mtime == 1} {
		dict set fid mtime $s(mtime)
	}

	dict set fid rehome $rehome
	dict set fid modbase $modbase

	return $fid
}

#
# Writes a Thatch header
#
proc fid->fic {of fid} {
	puts -nonewline $of [^Z][sig][fid->fis $fid]\n
}

#
# Writes a File Info Chunk (fic): the Thatch header and the file
# from a File Spec (fs)
#
proc fs->fic {of fs} {
	set fid [fs->fid $fs data]
	fid->fic $of $fid
	puts -nonewline $of $data
}

#
# Arguments (ags) -> File Spec List (fsl)
#
proc ags->fsl {ags} {
	set fsl [list]
	#         name      var     val default
	set al [list {*}{
		{ -         doArgs    0  1 }
		{ allsize   allSize   1 -1 }
		{ allmtime  allMtime  1 -1 }
		{ allbase64 allBase64 1 -1 }
		{ allzip    allZip    1 -1 }
		{ allmd5    allMd5    1 -1 }
		{ allsha256 allSha256 1 -1 }
		{ noallsize   allSize   0 -1 }
		{ noallmtime  allMtime  0 -1 }
		{ noallbase64 allBase64 0 -1 }
		{ noallzip    allZip    0 -1 }
		{ noallmd5    allMd5    0 -1 }
		{ noallsha256 allSha256 0 -1 }
		{ size      size      1 -1 }
		{ mtime     mtime     1 -1 }
		{ base64    base64    1 -1 }
		{ zip       zip       1 -1 }
		{ md5       md5       1 -1 }
		{ sha256    sha256    1 -1 }
		{ nosize    size      0 -1 }
		{ nomtime   mtime     0 -1 }
		{ nobase64  base64    0 -1 }
		{ nozip     zip       0 -1 }
		{ nomd5     md5       0 -1 }
		{ nosha256  sha256    0 -1 }
		{ allremove allRemove 1  0 }
		{ remove    remove    1 -1 }
		{ file      argIsFs   1  0 }
		{ home      argIsHm   1  0 }
		{ rehome    argIsRHm  1  0 }
		{ modbase   argIsMb   1  0 }
		{ rename    argIsRen  1  0 }
	}]
	foreach a $al { set [lindex $a 1] [lindex $a 3] }

	set    homedir ""
	set  rehomedir ""
	set modbasedir ""
	set resetOpts 1
	set gopts {}
	# Arg options override FileSpec options
	foreach arg $ags {
		if {$resetOpts} {
			set resetOpts 0
			set rename ""
			foreach a $al {
				# Only reset specific options
				if {[lindex $a 0] ni [fsFields]} { continue }
				set [lindex $a 1] [lindex $a 3]
			}
		}
		set home    ""
		set rehome  ""
		set modbase ""
		if {$argIsHm} {
			set argIsHm 0
			set homedir $arg
			continue
		}
		if {$argIsRHm} {
			set argIsRHm 0
			set rehomedir $arg
			continue
		}
		if {$argIsMb} {
			set argIsMb 0
			set modbasedir $arg
			continue
		}
		if {$argIsRen} {
			set argIsRen 0
			set rename $arg
			continue
		}
		# Does arg look like an option or a FileSpec?
		if {$doArgs && !$argIsFs && [string index $arg 0] eq "-"} {
			# Valid options are assigned to vars and added to gopts
			if {[set i [lsearch -index 0 $al [string range $arg 1 end]]] != -1} {
				set [lindex $al $i 1] [lindex $al $i 2]
				if {[string range $arg 1 3] eq "all" || [string range $arg 1 5] eq "noall"} {
					lappend gopts [string range $arg 1 end]
				}
				continue
			}
		}
		set argIsFs 0
		# Check FileSpec for options
		set opts [lassign [split $arg %] name]
		foreach opt $opts {
			if {$argIsHm} {
				set argIsHm 0
				set home $opt
				continue
			}
			if {($opt ni $gopts || $opt eq "home") && [set i [lsearch -index 0 $al $opt]] != -1} {
				set [lindex $al $i 1] [lindex $al $i 2]
			}
		}

		# Empty name can be used to set global options
		if {$name eq ""} {
			set resetOpts 1
			continue
		}

		# Use defaults if options not supplied
		#foreach n {size mtime base64 zip md5 sha256 remove rehome modbase} {}
		foreach n {size mtime base64 zip md5 sha256 remove} {
			if {[set $n] == -1} { set $n [set all[string totitle $n]] }
		}

		if {$home eq ""} {
			set home [expr {[string index $name 0] eq "/" ? "" : $homedir}]
		}

		set file [file join $home $name]

		if {$rename ne ""} {
			set name $rename
		}
		if {$rehomedir ne ""} {
			set rehome [file join [expr {[string index $name 0] eq "/" ? "" : $rehomedir}] $name]
		}
		if {$modbasedir ne "" && [regexp -- {^([_[:alpha:]][:_[:alnum:]]*)-([[:digit:]].*)[.]tm$} [file tail $file] -> pn pv]} {
			set modbase $modbasedir
		}

		lappend fsl [list $file $size $mtime $base64 $zip $md5 $sha256 $remove $rehome $modbase]

		set resetOpts 1
	}

	return $fsl
}

proc thatchOpen {fn {dir {in}}} {
	if {$dir eq "out"} {
		set mode w
		set eofc ""
	} else {
		set mode r
		set eofc [^Z]
	}
	return [openIt[string totitle $dir] $fn $mode -eofchar $eofc -encoding utf-8 -translation lf]
}

#
# Create or modify a Thatch file
#
proc thatchOp {op ifn ofn args} {
	set if [openItIn $ifn]
	if {$op eq "create"} {
		chan configure $if -eofchar [^Z]
	} else {
		chan configure $if -eofchar ""
	}
	set of [thatchOpen $ofn out]
	chan copy $if $of
	closeIt $if
	foreach fs [ags->fsl $args] {
		fs->fic $of $fs
	}
	closeIt $of
	return 1
}

proc thatchCopyCallbackMain {t fn f fid args} {
	lassign $args fsl of
	puts -nonewline $of [read $f]
	return skip
}

proc thatchCopyCallbackFile {t fn f fid args} {
	lassign $args fsl of

	set dosize 0
	set dobase64 0
	set undobase64 0
	set dozip 0
	set undozip 0
	set ofid $fid
	if {[set i [lsearch -index 0 $fsl [dict get $fid name]]] != -1} {
		if {[lindex $fsl $i [lsearch [fsFields] remove]]} {
			return ""
		}
		foreach fld [list size mtime] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0 && [dict exists $fid $fld]} {
				set fid [dict remove $fid $fld]
			} elseif {$v == 1} {
				if {$fld eq "size"} {
					set dosize 1
				}
			}
		}
		foreach fld [list base64 zip] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0} {
				if {![dict exists $fid transforms]} { continue }
				set transforms [dict get $fid transforms]
				if {[set i [lsearch $transforms $fld]] != -1} {
					set transforms [lreplace $transforms $i $i]
				}
				if {$fld eq "base64"} {set undobase64 1}
				if {$fld eq "zip"}    {set undozip 1}
			} elseif {$v == 1} {
				if {[dict exists $fid transforms]} { continue }
				lappend transforms $fld
				if {$fld eq "base64"} {set dobase64 1}
				if {$fld eq "zip"}    {set dozip 1}
			}
			if {[llength $transforms] == 0} {
				set fid [dict remove $fid transforms]
			} else {
				dict set fid transforms $transforms
			}
		}
		set checksums [dgwd {} $fid checksums]
		set sums1 [split $checksums :,]
		foreach fld [list md5 sha256] {
			set v [lindex $fsl $i [lsearch [fsFields] $fld]]
			if {$v == -1} { continue }
			if {$v == 0} {
				if {[set i [lsearch $sums1 $fld]] != -1} {
					set sums1 [lreplace $sums1 $i $i+1]
				}
			} elseif {$v == 1} {
			}
		}
		if {[llength $sums1] == 0} {
			if {[dict exists $fid checksums]} {
				set fid [dict remove $fid checksums]
			}
		} else {
			#dict set fid checksums something
		}
	}

	set d [read $f]

	if {$undobase64} {
		set d [thatchTransform $d decode {*}[dgwd {} $ofid transforms]]
		if {[dict exists $fid stored-size]} {
			set fid [dict remove $fid stored-size]
		}
	}
	if {$undozip} {
		set d [thatchTransform $d decode {*}[dgwd {} $ofid transforms]]
		if {[dict exists $fid stored-size]} {
			set fid [dict remove $fid stored-size]
		}
	}
	if {$dosize} {
		dict set fid size [string length $d]
	}
	if {$dobase64} {
		set d [thatchTransform $d encode {*}[dict get $fid transforms]]
		dict set fid stored-size [string length $d]
	}
	if {$dozip} {
		set d [thatchTransform $d encode {*}[dict get $fid transforms]]
		dict set fid stored-size [string length $d]
	}
	fid->fic $of [dict remove $fid offset]
	puts -nonewline $of $d

	return skip
}

proc thatchCopy {ifn ofn args} {
	set of [thatchOpen $ofn out]
	thatchTraverse $ifn [list [imn thatchCopyCallbackFile] [imn thatchCopyCallbackMain]] [ags->fsl $args] $of
	closeIt $of
	return 1
}

proc thatchNameMatch {fid g} {
	return [string match $g [dict get $fid name]]
}

proc zthatchFullFileGlobCallback {t fn f fid args} {
	variable res
	if {$t == 2} {
		dict set res main [read $f]
		return skip
	}
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res thIndex [dict get $fid name] $fid
		}
	}
	return ""
}

proc zthatchFullFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFullFileGlobCallback]] {*}$args
	return $res
}

proc thatchFullFileGlobCallback {t fn f fid args} {
	variable res
	if {$t == 2} {
		set size [string length [read $f]]
		dict set res $fn [dict set fid size $size]
		return skip
	}
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res [dict get $fid name] $fid
		}
	}
	return ""
}

proc thatchFullFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFullFileGlobCallback]] {*}$args
	return $res
}

proc thatchFileGlobCallback {t fn f fid args} {
	if {$t == 2} { return "" }
	variable res
	foreach a $args {
		if {[thatchNameMatch $fid $a]} {
			dict set res [dict get $fid name] $fid
		}
	}
	return ""
}

proc thatchFileGlob {fn args} {
	if {[llength $args] == 0} { set args [list *] }
	variable res
	set res [dict create]
	thatchTraverse $fn [list [imn thatchFileGlobCallback]] {*}$args
	return $res
}

proc thatchTraverse {fn cb args} {
	# This is supposed to be a Thatch file so it should conform
	set f [thatchOpen $fn]

	# Setup callbacks
	lassign $cb cb mcb
	if {$mcb eq ""} {
		set mcb $cb
	}

	# Get the 'main'
	set fid [dict create name $fn offset 0]
	set r [$mcb 2 $fn $f $fid {*}$args]
	if {$r ne "skip"} {
		read $f
	}

	# Note its largesse
	set offset [tell $f]

	# Now let's see if there's anything else
	while {1} {
		# We have hit eof so check if it's a ^Z eof or a real eof.
		# Do this by turning off -eofchar detection and reading one char:
		# it should be a ^Z. Set -eofchar back to ^Z after.
		fconfigure $f -eofchar {}
		set cz [read $f 1]
		# Real eof? If so, $cz should also be "".
		if {[eof $f] || $cz eq ""} {
			# Normal case
			if {[eof $f] && $cz eq ""} {
				break
			}
			# todo, strange case
			break
		}
		# Gotta be control-zee
		if {$cz ne [^Z]} { binary scan $cz c c; error "Expected ^Z at $offset, got \"[format %#x $c]\"::[gets $f]" }
		fconfigure $f -eofchar [^Z]
		# After the ^Z should be the sig
		set something [read $f [string length [sig]]]
		if {$something ne [sig]} {
			if {[eof $f]} {
				if {$something eq ""} {
					# null block
					continue
				}
			} else {
				append something [read $f]
			}
			# anon block
			continue
		}
		# Now we should be at the start of the file info line
		set fis [gets $f]
		# Do some checks ...
		if {[eof $f]} { explode }

		# Recalc offset:
		#    ^Z           \n                                 fis                               \n
		incr offset; incr offset [string length [sig]]; incr offset [string length $fis]; incr offset

		set fid [fis->fid $fis]
		dict set fid offset $offset
		set r [$cb 1 $fn $f $fid {*}$args]
		if {$r eq "stop"} { break }
		set size  [dgwd -1 $fid size]
		set ssize [dgwd -1 $fid stored-size]
		if {$ssize > 0} { set size $ssize }
		if {$size == -1} {
			if {$r ne "skip"} {
				read $f
			}
			set offset [tell $f]
		} else {
			if {$r ne "skip"} {
				seek $f $size current
			}
			incr offset $size
		}
	}
	closeIt $f
}

proc thatchExtract {fn fid} {
	set f [thatchOpen $fn]

	set name [lindex [dict keys $fid] 0]
	seek $f [dict get $fid $name offset] start

	set size [dgwd -1 $fid $name size]
	set ssize [dgwd -1 $fid $name stored-size]

	set transforms [dgwd {} $fid $name transforms]
	set checksums [dgwd "" $fid $name checksums]

	if {$ssize < 1} {
		set ssize $size
	}

	if {$ssize > 0} {
		set data [read $f $ssize]
	} else {
		set data [read $f]
	}

	closeIt $f

	set data [thatchTransform $data decode {*}[lreverse $transforms]]

	set checksum_errs [thatchCheckChecksums $data {*}$checksums]
	if {[llength $checksum_errs] > 0} {
		error "checksum mismatches ($fid):\n![join $checksum_errs \n]!"
	}

	return $data
}

proc thatchExtractMain {fn} {
	set f [thatchOpen $fn]
	set data [read $f]
	closeIt $f
	return $data
}

proc thatchIndexGlob {fn args} {
	variable thIndex
	if {[llength $args] == 0} { set args [list *] }
	set res [dict create]
	foreach a $args {
		set res [dict merge $res [dict filter $thIndex key $a]]
	}
	return $res
}

proc thatchCreateIndex {fn} {
	thatchDestroyIndex
	variable thIndex [thatchFileGlob $fn]
	interp alias {} [imn thatchGlob] {} [imn thatchIndexGlob]
}

proc thatchDestroyIndex {} {
	variable thIndex
	unset -nocomplain thIndex
	interp alias {} [imn thatchGlob] {} [imn thatchFileGlob]
}

#-------------------------------------------------------------------
# Overrides
#-------------------------------------------------------------------

proc thatchSourceOverride {fn args} {
	if {[dict size [set fid [thatchGlob $fn {*}$args]]] > 0} {
		set  script [info script]
		info script [lindex [dict keys $fid]]
		set  retval [uplevel 1 [thatchExtract $fn $fid]]]
		info script $script
	} else {
		set retval [uplevel 1 [list ::tcl::tcl_source {*}$args]]
	}
	return $retval
}

proc thatchImageOverride {fn args} {
	if {
		   [llength $args] > 3
		&& [lindex $args 0] eq "create"
		&& [set ind [lsearch $args -file]] != -1
		&& [dict size [set fid [thatchGlob $fn [lindex $args $ind+1]]]] > 0
	} {
		lset args $ind -data
		lset args $ind+1 [thatchExtract $fn $fid]
	}
	return [uplevel 1 [list ::tcl::tcl_image {*}$args]]
}

proc thatchLoadOverride {fn args} {
	set lfi -1
	foreach lfn $args {
		incr lfi
		if {[string index $lfn 0] ne "-"} {
			break
		}
	}
	if {$lfi > -1 && [dict size [set fid [thatchGlob $fn $lfn]]] > 0} {
		set c [file tempfile t thatch]
		chan configure $c -translation binary
		puts -nonewline $c [thatchExtract $fn $fid]
		close $c
		lset args $lfi $t
		if {[llength $args] > ($lfi - 1)} {
			lappend args [string totitle [lindex [regexp -inline {^lib([a-zA-z_]+).*} [file tail $lfn]] end]]
		}
		set code [catch {uplevel 1 [list ::tcl::tcl_load {*}$args]} res]
		file delete -force $t
	} else {
		set code [catch {uplevel 1 [list ::tcl::tcl_load {*}$args]} res]
	}
	return -code $code $res
}

proc thatchFileOverride {fn args} {
	set ok 0
	if {[llength $args] == 2} {
		set fcmd [lindex $args 0]
		if {$fcmd in {exists isfile owned readable executable isdirectory writable atime mtime attributes size type}} {
			set ok 1
		}
	}
	if {[llength $args] == 3} {
		set fcmd [lindex $args 0]
		if {$fcmd in {lstat stat}} {
			set ok 1
		}
	}

	if {[lindex $args 0] eq "type" && [lindex $args 1] eq ""} { return "directory" }
	if {$ok && [dict size [set fid [thatchGlob $fn [lindex $args 1]]]] == 1} {
		switch -exact -- [lindex $args 0] {
		exists -
		isfile -
		owned -
		readable {
			return 1
		}
		executable -
		isdirectory -
		writable {
			return 0
		}
		type {
			return "file"
		}
		atime -
		mtime {
			return [dict get $fid mtime]
		}
		attributes {
			return 0o444
		}
		size {
			return $size
		}
		lstat -
		stat {
			upvar 1 [lindex $args 2] a
			unset -nocomplain a
			array set a [list atime [dict get $fid mtime] ctime [dict get $fid mtime] dev -1 gid -1 ino -1 mode 0o444 mtime [dict get $fid mtime] nlink 1 size [dict get $fid size] type "file" uid -1]
			return ""
		}
		default { explode! }
		}
	}
	return [uplevel 1 [list ::tcl::tcl_file {*}$args]]
}

proc thatchGlobOverride {fn args} {
#puts $fn:$args
#if {[catch {info level -1} e]} { set e :: }
#puts $e:$args
	variable thIndex
	array set g {-join 0 -tails 0 -types {} -directory "" -path "" patterns {}}
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -- $arg {
			-join - -tails { set g($arg) 1 }
			-dir - -directory - -path - -types {
				if {$arg eq "-dir"} { set arg "-directory" }
				set state setelem
				set elem $arg
			}
			-nocomplain {}
			-- { set state endopts }
			default { lappend g(patterns) $arg }
			}
		}
		endopts { lappend g(patterns) $arg }
		setelem { set state opt; set g($elem) $arg }
		}
	}
	if {$g(-join)} { set g(patterns) [file join {*}$g(patterns)] }
	set fid [dict create]
	foreach p $g(patterns) {
		set fid [dict merge $fid [thatchGlob $fn [file join $g(-directory) $g(-path) $p]]]
	}
	if {[dict size $fid] > 0} {
		return [dict keys $fid]
	}
	tailcall ::tcl::tcl_glob {*}$args
}

#-------------------------------------------------------------------
# Setup
#-------------------------------------------------------------------

proc overrideAProc {fn proc} {
	if {[llength [info commands ::$proc]]} {
		rename ::$proc ::tcl::tcl_$proc
		interp alias {} ::$proc {} [imn thatch[string totitle $proc]Override] $fn
	} else {
		
	}
}

proc thatchSetup {fn args} {
	set oo {}; foreach o [set op {file glob image load source index packages modules}] { lappend oo -no$o }
	foreach a $args {
		if {[lsearch -exact $oo $a] == -1} {
			error [unknownOption $a $oo]
		}
		set i [lsearch -exact $op [string range $a 3 end]]
		set op [lreplace $op $i $i]
	}
	set createIndex 0
	if {[set i [lsearch -exact $op index]] != -1} {
		set op [lreplace $op $i $i]
		set createIndex 1
	}
	set setupPackages 0
	if {[set i [lsearch -exact $op packages]] != -1} {
		set op [lreplace $op $i $i]
		set setupPackages 1
	}
	set setupModules 0
	if {[set i [lsearch -exact $op modules]] != -1} {
		set op [lreplace $op $i $i]
		set setupModules 1
	}
	foreach p $op { overrideAProc $fn $p }
	if {$createIndex} { thatchCreateIndex $fn }
	if {$setupPackages} {
		dict for {k v} [thatchGlob $fn *pkgIndex.tcl *tclIndex] {
			set ::dir [file dirname $k]
			uplevel #0 [thatchExtract $fn [dict create $k $v]]
		}
		unset -nocomplain ::dir
	}
	if {$setupModules} {
		dict for {k v} [thatchGlob $fn {*-[0-9]*.tm}] {
			if {![regexp -- {^([_[:alpha:]][:_[:alnum:]]*)-([[:digit:]].*)[.]tm$} [file tail $k] -> pn pv]} { continue }
			if {[dict exists $v modbase]} {
				set pn [string map {/ ::} [string trimleft [string range $k [string length [dict get $v modbase]] end-[string length [file tail $k]]] /]]$pn
			}
			package ifneeded $pn $pv "[::list package provide $pn $pv];[::list source -encoding utf-8 $k]"
		}
	}
}

#-------------------------------------------------------------------
# Package Setup
#-------------------------------------------------------------------

proc setup {} {
	package require Tcl 8.5
	aliasconst ^Z \x1a sig "\n# THATCH "
	interp alias {} [imn thatchCreate] {} [imn thatchOp] create
	interp alias {} [imn thatchAdd]    {} [imn thatchOp] add
	thatchDestroyIndex
	namespace export thatch*
	package provide thatch 0.5
}

namespace current}]::setup

# EOF
